from openobd.core.stream_handler import StreamHandler
from openobd.core.session import OpenOBDSession
import logging


def set_bus_configuration(openobd_session: OpenOBDSession, bus_configs):
    # Open a configureBus stream, send the bus configurations, and close the stream
    bus_config_stream = StreamHandler(openobd_session.configure_bus)
    bus_config_stream.send_and_close(bus_configs)
    logging.debug("Buses have been configured.")




from typing import Union

from google.protobuf.message import Message
from openobd_protocol.Function.Messages.Function_pb2 import VariableList, ContextType

from openobd.core.exceptions import OpenOBDException
from openobd.functions.message import pack, unpack_variable_list, unpack_variable
from openobd.core.session import OpenOBDSession
from collections.abc import MutableMapping

class ContextVariables(MutableMapping):

    openobd_session = None
    context_type = None
    message_classes = None

    def __init__(self, openobd_session: OpenOBDSession,
                 context_type: ContextType = ContextType.FUNCTION_CONTEXT, message_classes = None):

        self.openobd_session = openobd_session
        self.context_type = context_type
        self.message_classes = message_classes if message_classes is not None else []

    def __getitem__(self, key: str) -> Union[str, Message, None]:
        try:
            variable = self.openobd_session.get_context_variable(key=key, context_type=self.context_type)
        except OpenOBDException as e:
            '''gRPC standard defines status code 5 as NOT_FOUND'''
            '''If the key does not exist, just return None'''
            if e.status == 5:
                return None

            '''Otherwise raise the exception'''
            raise e

        return unpack_variable(variable, self.message_classes)

    def __setitem__(self, key: str, value: Union[str, Message]):
        self.openobd_session.set_context_variable(key=key, value=pack(value),
                                                  context_type=self.context_type)

    def __delitem__(self, key: str):
        self.openobd_session.delete_context_variable(key=key,
                                                     context_type=self.context_type)

    def get_variables(self, prefix=""):
        variable_list = (self.openobd_session
                         .get_context_variable_list(prefix=prefix, context_type=self.context_type)) # type: VariableList

        return unpack_variable_list(variable_list, self.message_classes)

    def __iter__(self, prefix=""):
        return iter(self.get_variables(prefix))

    def __len__(self, prefix=""):
        return len(self.get_variables(prefix))

    def __str__(self):
        return str(self.get_variables())



import logging
from logging import LogRecord
from typing import Optional

from colorlog import ColoredFormatter

from openobd.core.session import OpenOBDSession
from openobd.core.stream_handler import StreamHandler
from openobd_protocol.Logging.Messages.LogMessage_pb2 import LogMessage, LogLevel

def initialize_logging(log_level: Optional[str]):
    if log_level is None or not isinstance(log_level, str):
        log_level = "INFO"

    handler = logging.StreamHandler()
    formatter = ColoredFormatter(
        "%(log_color)s%(levelname)-8s%(reset)s %(blue)s%(threadName)s: %(white)s%(message)s",
        log_colors={
            'DEBUG': 'cyan',
            'INFO': 'green',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'bold_red',
        }
    )
    handler.setFormatter(formatter)
    logging.basicConfig(handlers=[handler], level=log_level.upper())


class ScriptLogHandler(logging.Handler):
    """
    A class which sends records to an openOBD logging stream. Please note that
    formatting does not have effect on this handler. Messages are sent as-is to the engine
    and are formatted in the engine according to the rules defined server-side. This ensures a
    consistent interface for the operator.
    """
    def __init__(self, session: OpenOBDSession, level=logging.NOTSET):
        super().__init__(level)
        self.stream_handler = StreamHandler(session.open_log_stream, outgoing_stream=True)

    def close(self):
        super().close()
        self.stream_handler.stop_stream()

    def emit(self, record: LogRecord):
        self.stream_handler.send(LogMessage(
            content = record.getMessage(),
            level = self._get_grpc_level_from_name(record.levelname)
        ))

    def _get_grpc_level_from_name(self, level_name: str) -> LogLevel:
        match level_name:
            case 'CRITICAL' | 'ERROR':
                return LogLevel.LOG_LEVEL_ERROR
            case 'WARNING':
                return LogLevel.LOG_LEVEL_WARNING
            case 'DEBUG':
                return LogLevel.LOG_LEVEL_DEBUG
            case 'INFO' | 'NOTSET' | _:
                return LogLevel.LOG_LEVEL_INFO



import inspect

# from google.protobuf.internal.well_known_types import Any
from google.protobuf.message import Message
from openobd_protocol.Function.Messages.Function_pb2 import VariableList, Variable
from google.protobuf.any_pb2 import Any


def is_protobuf(value):
    return isinstance(value, Message) or (isinstance(value, type) and issubclass(value, Message))

def is_packed(value):
    return isinstance(value, Any)

def pack(value):
    if is_protobuf(value) and not is_packed(value):
        any = Any()
        any.Pack(value)
        return any
    return value

def unpack(value, message_classes = None):
    if not is_protobuf(value):
        return value
    if not is_packed(value):
        return value
    if message_classes is None:
        return value

    '''Try to find the data type'''
    for message_class in message_classes:
        '''Try to unpack with one of our known types'''
        msg_instance = message_class()
        if value.Unpack(msg_instance):
            return msg_instance

    return value

def unpack_variable(variable: Variable, message_classes = None):
    if variable.WhichOneof("Val") == "object":
        value = unpack(variable.object, message_classes)
    else:
        value = variable.value
    return value

def unpack_variable_list(variable_list: VariableList, message_classes = None):
    variables = {}
    for var in variable_list.variables:
        variables[var.key] = unpack_variable(var, message_classes)
    return variables

'''Inspect the signature of a function for protobuf Message types and return them as a list'''
def get_message_classes_from_function_signature(function):
    signature_classes = []
    sig = inspect.signature(function)
    for name, param in sig.parameters.items():
        if is_protobuf(param.annotation):
            '''If it is a protobuf message we will handle it for this function'''
            signature_classes.append(param.annotation)

    '''The return signature'''
    if is_protobuf(sig.return_annotation):
        signature_classes.append(sig.return_annotation)

    return signature_classes


import argparse
import os
import sys
import textwrap
import typing
from typing import Literal, Optional, List

from openobd import ConnectionArguments, SessionArguments
from openobd.core.arguments import ExecutorArguments, ArgumentStore

COMMAND_RUN = 'run'
COMMAND_SERVE = 'serve'

'''The openOBD environment variables that can be set'''
environment_variable_spec = {
    "client_id":        "OPENOBD_PARTNER_CLIENT_ID",
    "client_secret":    "OPENOBD_PARTNER_CLIENT_SECRET",
    "cluster_id":       "OPENOBD_CLUSTER_ID",
    "grpc_host":        "OPENOBD_GRPC_HOST",
    "grpc_port":        "OPENOBD_GRPC_PORT",
    "token":            "OPENOBD_TOKEN",
    "ticket_id":        "OPENOBD_TICKET_ID",
    "connector_id":     "OPENOBD_CONNECTOR_ID",
    "log_level":        "OPENOBD_LOG_LEVEL",
}

_program = 'python -m openobd'

class EnvironmentVariableParser:
    _constructor_args: dict[str, typing.Any] = {}

    def __init__(self,
                 # openOBD gRPC server credentials
                 client_id: str = None,
                 client_secret: str = None,
                 cluster_id: str = None,
                 grpc_host: str = None,
                 grpc_port: int = None,

                 # Load the following modules with openOBD scripts
                 modules: list[str] = None,
                 unique: bool = None,
                 prefix: str = None,

                 bypass_function_broker: bool = None,
                 ticket_id: str = None,
                 connector_id: str = None,
                 token: str = None,

                 log_level: Optional[str] = None
                 ):
        """
        Initialize arguments according to the following priorities:
         - keyword argument, if not available we use:
         - command line argument, if not available we use:
         - environment variable


        :param client_id: Client id of the Jifeline Partner (Partner API credentials)
        :param client_secret: Client secret of the Jifeline Partner (Partner API credentials)
        :param cluster_id: Server cluster ('001' for Europe, '002' for USA)
        :param grpc_host: gRPC host of the openOBD service (default is 'grpc.openobd.com')
        :param grpc_port: gRPC port of the openOBD service (default is 443)
        :param modules: List of modules (paths to openOBD scripts) that need to be initialized
        :param unique: Generate a fresh openOBD script id every time we initialize the scripts (Development)
        :param prefix: Prefix the script names with a short string (max 4 characters) (Development)
        :param bypass_function_broker: Calls to other openOBD functions are not routed through the function broker when
            they are locally known.
        :param ticket_id: Create an openOBD session based on a ticket
        :param connector_id: Create an openOBD session based on a connector (Development)
        :param token: Create an openOBD session based in an authentication token
        """
        self._constructor_args["client_id"] = client_id
        self._constructor_args["client_secret"] = client_secret
        self._constructor_args["cluster_id"] = cluster_id
        self._constructor_args["host"] = grpc_host
        self._constructor_args["port"] = grpc_port

        self._constructor_args["modules"] = modules
        self._constructor_args["unique"] = unique
        self._constructor_args["prefix"] = prefix

        self._constructor_args["ticket_id"] = ticket_id
        self._constructor_args["connector_id"] = connector_id
        self._constructor_args["token"] = token
        self._constructor_args["bypass_function_broker"] = bypass_function_broker

        self._constructor_args["log_level"] = log_level

    def get_from_env_with_default(self, key: str, default: Optional[typing.Any] = None):
        if key in self._constructor_args and self._constructor_args[key] is not None:
            return self._constructor_args[key]

        if key in environment_variable_spec:
            env_key = environment_variable_spec[key]
            if env_key is not None and env_key in os.environ and os.environ[env_key] is not None:
                return os.environ[env_key]

        return default

    def parse(self) -> ArgumentStore:
        return ArgumentStore(
            session_arguments = SessionArguments(
                ticket=self.get_from_env_with_default("ticket_id"),
                token=self.get_from_env_with_default("token"),
                connector=self.get_from_env_with_default("connector_id"),
                bypass_function_broker=self.get_from_env_with_default("function", False)
            ),
            connection_arguments = ConnectionArguments(
                client_id = self.get_from_env_with_default("client_id"),
                client_secret = self.get_from_env_with_default("client_secret"),
                cluster_id = self.get_from_env_with_default("cluster_id"),
                grpc_host = self.get_from_env_with_default("grpc_host"),
                grpc_port = self.get_from_env_with_default("grpc_port")
            ),
            executor_arguments = ExecutorArguments(
                modules = self.get_from_env_with_default("modules", []),
                unique_function_ids = self.get_from_env_with_default("unique", False),
                prefix = self.get_from_env_with_default("prefix")
            ),
            log_level = self.get_from_env_with_default("log_level")
        )


from openobd.functions.function import OpenOBDFunction
from openobd.core.exceptions import OpenOBDException

class OpenOBDComposition(OpenOBDFunction):

    def __enter__(self):
        if self._context_finished_:
            ''' We cannot continue when the constructor failed '''
            raise OpenOBDException("Failed to construct openOBD session")

        return self





import os
import re
from typing import Optional, Any

def redacted_str(obj: Any, redacted_attributes: list[str] = []) -> str:
    props = [prop for prop in dir(obj) if not prop.startswith('__') and not callable(getattr(obj, prop))]
    return (f"{type(obj)}(" +
            [f"{prop}={'***' if prop in redacted_attributes else getattr(obj, prop)}" for prop in props] +
            ")")

class ExecutorArguments:
    modules: list[str] = None
    unique_function_ids: bool = None
    prefix: str = None

    def __init__(self,
                 # Load the following modules with openOBD scripts
                 modules: list[str] = None,
                 unique_function_ids: bool = None,
                 prefix: str = None):
        """
        :param modules: List of modules (paths to openOBD scripts) that need to be initialized
        :param unique_function_ids: Generate a fresh openOBD script id every time we initialize the scripts (Development)
        :param prefix: Prefix the script names with a short string (max 4 characters) (Development)
        """
        self.modules = modules
        self.unique_function_ids = unique_function_ids
        self.prefix = self._filter_prefix(prefix)

    @staticmethod
    def _filter_prefix(prefix: Optional[str]):
        if prefix is None:
            return None

        '''Only alphanumeric'''
        filtered = re.sub(r'[^a-zA-Z0-9]', '', prefix)
        '''Max 4 chars'''
        return filtered[:4].upper()

    def __str__(self):
        return redacted_str(self)

class SessionArguments:
    token: Optional[str]
    ticket_id: Optional[str]
    connector_id: Optional[str]
    bypass_function_broker: bool = False

    def __init__(self,
                 # Specific 'run' arguments
                 ticket: Optional[str] = None,
                 connector: Optional[str] = None,
                 token: Optional[str] = None,
                 bypass_function_broker: bool = False):
        """
        :param ticket: Create an openOBD session based on a ticket
        :param connector: Create an openOBD session based on a connector (Development)
        :param token: Create an openOBD session based in an authentication token
        :param bypass_function_broker: Calls to other openOBD functions are not routed through the function broker when
            they are locally known.
        """
        self.ticket_id = ticket
        self.connector_id = connector
        self.token = token
        self.bypass_function_broker = bypass_function_broker

    def __str__(self):
        return redacted_str(self)

class ConnectionArguments:
    client_id: str
    client_secret: str
    cluster_id: str
    grpc_host: str
    grpc_port: int

    def __init__(self,
                 # openOBD gRPC server credentials
                 client_id: str = None,
                 client_secret: str = None,
                 cluster_id: str = None,
                 grpc_host: str = None,
                 grpc_port: int = None,
                 ):
        """
        :param client_id: Client id of the Jifeline Partner (Partner API credentials)
        :param client_secret: Client secret of the Jifeline Partner (Partner API credentials)
        :param cluster_id: Server cluster ('001' for Europe, '002' for USA)
        :param grpc_host: gRPC host of the openOBD service (default is 'grpc.openobd.com')
        :param grpc_port: gRPC port of the openOBD service (default is 443)
        """
        self.client_id = self._require_argument('client_id', client_id)
        self.client_secret = self._require_argument('client_secret', client_secret)
        self.grpc_host = grpc_host if grpc_host is not None else 'grpc.openobd.com'
        self.grpc_port = grpc_port if grpc_port is not None else 443
        self.cluster_id = self._require_argument('cluster_id', cluster_id)

    # Convenience method used to fetch arguments from environment variables
    # In older versions of the library, the environment variables were the only
    # way to provide parameters. Some places still require this setup (as a default)
    @staticmethod
    def from_environment_variables() -> "ConnectionArguments":
        return ConnectionArguments(
            client_id = os.environ.get("OPENOBD_PARTNER_CLIENT_ID"),
            client_secret = os.environ.get("OPENOBD_PARTNER_CLIENT_SECRET"),
            cluster_id = os.environ.get("OPENOBD_CLUSTER_ID"),
            grpc_host = os.environ.get("OPENOBD_GRPC_HOST"),
            grpc_port = os.environ.get("OPENOBD_GRPC_PORT"),
        )

    def _require_argument(self, argument: str, value):
        if value is None:
            message = f"Parameter '{argument}' is required but could not be found."
            raise AssertionError(message)

        return value

    def __str__(self):
        return redacted_str(self, ['client_secret'])

class ArgumentStore:
    connection_arguments: ConnectionArguments
    session_arguments: SessionArguments
    executor_arguments: ExecutorArguments

    log_level: str = 'INFO'

    def __init__(self,
                 connection_arguments: ConnectionArguments,
                 session_arguments: SessionArguments,
                 executor_arguments: ExecutorArguments,
                 log_level: str = 'INFO',
                 ):
        self.connection_arguments = connection_arguments
        self.session_arguments = session_arguments
        self.executor_arguments = executor_arguments
        self.log_level = log_level if log_level is not None else 'INFO'



import argparse
import os
import sys
import textwrap
from typing import Literal, Optional, List

from openobd import ConnectionArguments, SessionArguments
from openobd.core.arguments import ExecutorArguments, ArgumentStore
from openobd.environment_variables import environment_variable_spec

COMMAND_RUN = 'run'
COMMAND_SERVE = 'serve'

_program = 'python -m openobd'

class CliArgumentStore(ArgumentStore):
    file: str

    command: Literal['run', 'serve']

    def __init__(self,
                 command: Literal['run', 'serve'],

                 connection_arguments: ConnectionArguments,
                 session_arguments: SessionArguments,
                 executor_arguments: ExecutorArguments,

                 file: str,
                 log_level: str = 'INFO'):
        super().__init__(connection_arguments, session_arguments, executor_arguments, log_level)

        self.command = command
        self.file = file

    def is_command_run(self):
        return self.command == COMMAND_RUN

    def is_command_serve(self):
        return self.command == COMMAND_SERVE

class CliParser:
    # List of values as if they were set on the command line (e.g. [ '--prefix', 'abc', '--port', 80])
    _constructor_args: List[str] = []
    _command: Literal['run', 'serve'] = 'run'

    def __init__(self,
                 command: Optional[Literal['run', 'serve']] = None,

                 # openOBD gRPC server credentials
                 client_id: str = None,
                 client_secret: str = None,
                 cluster_id: str = None,
                 grpc_host: str = None,
                 grpc_port: int = None,

                 # Load the following modules with openOBD scripts
                 modules: list[str] = None,
                 unique: bool = None,
                 prefix: str = None,

                 # Specific 'run' arguments
                 file: str = None,
                 bypass_function_broker: bool = None,
                 ticket: str = None,
                 connector: str = None,
                 token: str = None,
                 ):
        """
        Initialize arguments according to the following priorities:
         - keyword argument, if not available we use:
         - command line argument, if not available we use:
         - environment variable

        :param command: Either 'run' or 'serve'.
                        'run': Runs an openOBD script
                        'serve': Serves openOBD scripts to the function broker
        :param client_id: Client id of the Jifeline Partner (Partner API credentials)
        :param client_secret: Client secret of the Jifeline Partner (Partner API credentials)
        :param cluster_id: Server cluster ('001' for Europe, '002' for USA)
        :param grpc_host: gRPC host of the openOBD service (default is 'grpc.openobd.com')
        :param grpc_port: gRPC port of the openOBD service (default is 443)
        :param modules: List of modules (paths to openOBD scripts) that need to be initialized
        :param unique: Generate a fresh openOBD script id every time we initialize the scripts (Development)
        :param prefix: Prefix the script names with a short string (max 4 characters) (Development)
        :param file: The file containing an openOBD script or composition that needs to be run
        :param bypass_function_broker: Calls to other openOBD functions are not routed through the function broker when
            they are locally known.
        :param ticket: Create an openOBD session based on a ticket
        :param connector: Create an openOBD session based on a connector (Development)
        :param token: Create an openOBD session based in an authentication token
        """
        self._command = command if command is not None else 'run'

        self._set_constructor_arg('--client-id', client_id)
        self._set_constructor_arg('--client-secret', client_secret)
        self._set_constructor_arg('--cluster-id', cluster_id)
        self._set_constructor_arg('--host', grpc_host)
        self._set_constructor_arg('--port', grpc_port)

        self._set_constructor_arg('--modules', modules)
        self._set_constructor_arg('--unique', unique)
        self._set_constructor_arg('--prefix', prefix)

        self._set_constructor_arg('--file', file)
        self._set_constructor_arg('--ticket', ticket)
        self._set_constructor_arg('--connector', connector)
        self._set_constructor_arg('--token', token)
        self._set_constructor_arg('--bypass-function-broker', bypass_function_broker)

    # def _set_file(self, file_path: str = None):
    #     if file_path is None:
    #         script_path = os.path.abspath(sys.argv[0])
    #         cwd = os.getcwd()
    #         file_path = os.path.relpath(script_path, cwd)

    '''Set values when defined in constructor'''
    def _set_constructor_arg(self, argument, value):
        if value is None:
            return

        self._constructor_args.append(argument)

        '''When the argument is a flag on the commandline we are done, just return'''
        if isinstance(value, bool):
            return

        '''When the argument is a list of values'''
        if isinstance(value, list):
            for val in value:
                self._constructor_args.append(val)
            return

        self._constructor_args.append(value)

    def parse(self) -> CliArgumentStore:
        # Build a parser with the value of the environment variables as the default
        parser = ArgumentParserFactory().build()

        # The command is supposed to be the first argument. If it is not given, use the default
        if len(sys.argv) == 1:
            cli_args = [self._command]
        else:
            cli_args = sys.argv[1:]

        # Use the constructor arguments and cli arguments as default
        default_args = cli_args + self._constructor_args

        # Parse the CLI arguments
        args = parser.parse_args(default_args)

        session_arguments = None
        file = None

        if args.command == COMMAND_RUN:
            session_arguments = SessionArguments(
                ticket=args.ticket_id,
                token=args.token,
                connector=args.connector_id,
                bypass_function_broker=args.bypass_function_broker
            )
            file = args.file if 'file' in args else None

        return CliArgumentStore(
            command = args.command,
            connection_arguments = ConnectionArguments(
                client_id = args.client_id,
                client_secret = args.client_secret,
                cluster_id = args.cluster_id,
                grpc_host = args.grpc_host,
                grpc_port = args.grpc_port
            ),
            session_arguments = session_arguments,
            executor_arguments = ExecutorArguments(
                modules = args.modules,
                unique_function_ids = args.unique if 'unique' in args else False,
                prefix = args.prefix if 'prefix' in args else None
            ),
            file = file,
            log_level = args.log_level
        )

class ArgumentParserFactory:
    def build(self):
        parser = argparse.ArgumentParser(add_help=False,
                                         prog=_program,
                                         description="Run openOBD on the remote diagnostics network of Jifeline Networks",
                                         epilog=f"For detailed help for a subcommand, run \n {os.path.basename(__file__)} <command> -h")

        subparsers = parser.add_subparsers(help='Command')
        subparsers.required = False
        subparsers.dest = 'command'

        self._add_command_run(subparsers)
        self._add_command_serve(subparsers)

        return parser

    '''Run command'''
    def _add_command_run(self, subparsers):
        parser_cmd_run = subparsers.add_parser('run',
                                               help='Run a openOBD function or composition',
                                               formatter_class=argparse.RawTextHelpFormatter,
                                               )

        description = textwrap.dedent('''\
            
            [ Run options ]
            ''')

        parent_group = parser_cmd_run.add_argument_group(description=description)
        parent_group.add_argument('--file', metavar='<filename>', dest="file", help='The path to a Python file. The file should contain a class based on the OpenOBDFunction or OpenOBDComposition class.',required=True)
        self._add_modules_arguments(parent_group)
        parent_group.add_argument('--bypass-function-broker', action='store_true', dest="bypass_function_broker", help='Bypass the function broker for functions that are locally available (loaded with --modules flag).')

        description=textwrap.dedent('''\
            
            [ Session options ]
            
              An openOBD session can be instantiated by
                     <ticket id>,
                OR   <connector id> (for development),
                OR   <token>.''')

        self._add_session_argument_group(parser_cmd_run, description)

        description=textwrap.dedent('''\
            
            [ openOBD server settings ]
            
              Required when session needs to be created or subsequent function calls
              need to be made (e.g. through the function broker when hosting functions).''')
        self._add_grpc_argument_group(parser_cmd_run, description)

        self._log_level_help(parser_cmd_run)

        description = textwrap.dedent('''\
    
            -------------------------------------------------------------------------------
            
            Example:
            
                {program} run --file example_functions/test.py --modules example_functions
                
                {program} run --file example_functions/test.py --ticket 8832507 --modules example_functions --bypass-function-broker
                
            '''.format(program=_program))

        parser_cmd_run.add_argument_group(description=description)


    '''Serve command'''
    def _add_command_serve(self, subparsers):
        '''Serve'''
        parser_cmd_serve = subparsers.add_parser('serve',
                                                 help='Host openOBD functions or compositions on the Jifeline network.',
                                                 formatter_class=argparse.RawTextHelpFormatter
                                                 )

        self._add_modules_arguments(parser_cmd_serve, add_prefix_option=True)

        description = textwrap.dedent('''\
    
            [ openOBD server settings ]
            
              Required to authenticate to the function broker.''')
        self._add_grpc_argument_group(parser_cmd_serve, description)

        self._log_level_help(parser_cmd_serve)

        description = textwrap.dedent('''\
    
            -------------------------------------------------------------------------------
    
            Example:
            
                {program} serve --modules example_functions
            '''.format(program=_program))

        parser_cmd_serve.add_argument_group(description=description)


    def _add_grpc_argument_group(self, parser, description: str):
        group = parser.add_argument_group(description=description)
        group.add_argument('--host', metavar='<grpc_host>', type=str,
                           dest="grpc_host",
                           default=os.environ.get(environment_variable_spec['grpc_host'], 'grpc.openobd.com'),
                           help=f"The gRPC host of the openOBD service (default is 'grpc.openobd.com'){self._env_help('grpc_host')}")

        group.add_argument('--port', metavar='<grpc_port>', type=int,
                           dest="grpc_port",
                           default=os.environ.get(environment_variable_spec['grpc_port'], "443"),
                           help=f"The gRPC port of the openOBD service (default is 443){self._env_help('grpc_port')}")

        group.add_argument('--cluster-id', metavar='<cluster_id>', type=str,
                           dest="cluster_id",
                           default=os.environ.get(environment_variable_spec['cluster_id'], "001"),
                           help=f"The cluster id of the Jifeline partner (001=EU, 002=USA, default is '001'){self._env_help('cluster_id')}")

        group.add_argument('--client-id', metavar='<client_id>', type=str,
                           dest="client_id",
                           default=os.environ.get(environment_variable_spec['client_id']),
                           help=f"The client id of the Jifeline partner.{self._env_help('client_id')}")

        group.add_argument('--client-secret', metavar='<client_secret>', type=str,
                           dest="client_secret",
                           default=os.environ.get(environment_variable_spec['client_secret']),
                           help=f"The client secret of the Jifeline partner.{self._env_help('client_secret')}")

        return parser

    def _add_session_argument_group(self, parser, description):
        group = parser.add_argument_group(description=description)
        group.add_argument('--ticket', metavar='<ticket id>', type=str,
                           dest='ticket_id',
                           default=os.environ.get(environment_variable_spec['ticket_id']),
                           help=f"Create an openOBD session on a specific ticket{self._env_help('ticket_id')}")

        group.add_argument('--connector', metavar='<connector id>', type=str,
                           dest='connector_id',
                           default=os.environ.get(environment_variable_spec['connector_id']),
                           help=f"Create an openOBD session using a specific connector (only for development){self._env_help('connector_id')}")

        group.add_argument('--token', metavar='<token>', type=str,
                           dest='token',
                           default=os.environ.get(environment_variable_spec['token']),
                           help=f"Authentication token of an available openOBD session{self._env_help('token')}")

        return parser

    def _add_modules_arguments(self, parser, add_prefix_option=False):
        parser.add_argument('--modules', dest="modules", nargs='+', metavar='<module>', default=None, help="Path to modules containing openOBD scripts (multiple paths can be provided)", required=False)
        parser.add_argument('--unique', action='store_true', dest="unique",
                            help='Generate fresh function ids for all functions that are loaded. Guarantees unique function registrations at the function broker.')
        if not add_prefix_option:
            return

        parser.add_argument('--prefix', metavar='<prefix>', dest="prefix", default=None, help='Prepend the name of every function with this prefix (max 4 alphanumeric characters and only in combination with --unique).',required=False)


    def _log_level_help(self, parser):
        log_help_group = parser.add_argument_group(description='[ Set log level ]')

        log_help = textwrap.dedent('''\
            Possible log levels:
             [CRITICAL, ERROR, WARNING, INFO, DEBUG] (Default is 'INFO')''')

        log_help_group.add_argument('--log-level', metavar='<log level>', type=str,
                                    dest='log_level',
                                    default=os.environ.get(environment_variable_spec['log_level']),
                                    help=f"{log_help}{self._env_help('log_level')}")

    @staticmethod
    def _env_help(arg_key: str):
        return f"\n - Environment variable: [{environment_variable_spec[arg_key]}]\n "






import copy
import sys
import typing

from openobd.core import GrpcFactory
from openobd.core.exceptions import OpenOBDException
from openobd.core.session import OpenOBDSession
from openobd.functions.function import OpenOBDFunction
from openobd.core.arguments import ConnectionArguments
from openobd_protocol.FunctionBroker.Messages.FunctionBroker_pb2 import FunctionUpdate, FunctionUpdateType, FunctionUpdateResponse, FunctionCall

from openobd.core.function_broker import OpenOBDFunctionBroker
from openobd.core.stream_handler import StreamHandler
from openobd.core.exceptions import OpenOBDStreamStoppedException

from openobd.functions.function_executor import FunctionExecutor

import threading
import logging

class FunctionLauncher:

    def __init__(self, function_broker: OpenOBDFunctionBroker, function_executor: FunctionExecutor, grpc_factory: typing.Optional[GrpcFactory] = None):
        """
        Handles communication with the function broker about functions that are available for execution.

        :param function_broker: Reference to a function broker instance
        :param function_executor: Functions that can be executed by the launcher
        """
        self.function_broker = function_broker
        self.function_executor = function_executor
        self.function_executor.load_modules(self.function_broker)
        self.stream_handler = None
        self.process_function_updates_thread = None
        self.grpc_factory = grpc_factory

        # List of threads with running functions, ordered by time of invocation
        self.threads = []

    def serve(self):
        """
        Starts serving functions to the function broker
        """
        function_registrations = self.function_executor.get_function_registrations()
        if len(function_registrations) == 0:
            logging.critical("No functions to serve!")
            logging.critical("A function should be defined according to the following pattern:\n" + '''
class ECUReset(OpenOBDFunction):

    version = "v0.1"
    name = "ECU Reset"
    description = "Reset ECU"

    def run(self):
        logging.info("Reset ECU")                        

            ''')
            sys.exit(1)

        '''Establish function update stream'''
        self.stream_handler = StreamHandler(self.function_broker.open_function_stream, outgoing_stream=True)

        for function_registration in function_registrations:
            self.stream_handler.send(FunctionUpdate(type=FunctionUpdateType.FUNCTION_UPDATE_TYPE_REQUEST, function_registration=function_registration))

        self.process_function_updates_thread = threading.Thread(target=self._process_function_updates,
                                                                name="Process function updates",
                                                                daemon=True)
        self.process_function_updates_thread.start()

        try:
            self.process_function_updates_thread.join()
        except KeyboardInterrupt as e:
            self._interrupt_launcher()

    def _process_function_updates(self):
        try:
            while True:
                function_update = self.stream_handler.receive()
                self._function_update(function_update)
        except OpenOBDStreamStoppedException:
            # This can happen when the function broker closes the stream due to no functions having been online for some time
            logging.warning("The stream to the function broker has been closed.")
        except OpenOBDException as e:
            logging.error(f"Stopped processing function updates due to an exception.")
            logging.error(e)
        finally:
            self.stream_handler.stop_stream()

    def _interrupt_launcher(self):
        # Loop over all active threads, starting with the one started most recently
        # This ensures that 'child' functions are always stopped before the 'parent' function
        for (function_reference, function_thread) in reversed(self.threads):
            if function_reference.__is_active__():
                logging.info(f"Stopping {function_reference}...")
                function_reference.interrupt()

        if self.stream_handler is not None:
            self.stream_handler.stop_stream()

    def _function_update(self, function_update: FunctionUpdate):
        if function_update.type == FunctionUpdateType.FUNCTION_UPDATE_TYPE_REQUEST:
            if function_update.HasField('function_call'):
                call = function_update.function_call
                logging.info(f"Received a function call for function id [{function_update.function_call.id}]")

                try:
                    # Test if function is available. Will raise if not available
                    self._available(call)

                    function_reference = self._get_function_instance(call)

                    function_thread = threading.Thread(target=self._call,
                                                     name=f"Function [{function_update.function_call.id}]",
                                                     args=[function_reference],
                                                     daemon=True)

                    self.threads.append((function_reference, function_thread))
                    function_thread.start()

                    # Immediately report back that we successfully received the function call and started the function
                    self.stream_handler.send(
                        FunctionUpdate(
                            type=FunctionUpdateType.FUNCTION_UPDATE_TYPE_RESPONSE,
                            function_call=function_update.function_call,
                            response=FunctionUpdateResponse.FUNCTION_UPDATE_SUCCESS
                        ))

                except Exception as e:
                    self.stream_handler.send(
                        FunctionUpdate(
                            type=FunctionUpdateType.FUNCTION_UPDATE_TYPE_RESPONSE,
                            function_call=function_update.function_call,
                            response=FunctionUpdateResponse.FUNCTION_UPDATE_FAILED,
                            response_description=str(e)
                        ))

            elif function_update.HasField('function_registration'):
                raise OpenOBDException("We do not expect a 'function_registration' as request from the function broker.")

            elif function_update.HasField('function_broker_token'):
                logging.info(f"Function update stream is still open")
                logging.debug(f"Received a new function broker token")

                # Acknowledge the ping/token update
                # We need to communicate to the broker every so often (before the Connection idle timeout of the Load Balancer)
                # or the stream (and all functions we host) gets closed
                response = copy.copy(function_update)
                response.type=FunctionUpdateType.FUNCTION_UPDATE_TYPE_RESPONSE
                self.stream_handler.send(response)

                logging.debug(f"Send new function broker token response")

            elif function_update.HasField('function_broker_reconnect'):
                raise OpenOBDException("Received a reconnect request from the function broker. We do not support it yet.")

            else:
                raise OpenOBDException("Unsupported request from the function broker")

        elif function_update.type == FunctionUpdateType.FUNCTION_UPDATE_TYPE_RESPONSE:
            logging.debug("Received function update response...")

            if function_update.HasField('function_registration') and function_update.HasField('response'):
                logging.debug(f"Received a function update for function id [{function_update.function_registration.details.id}]")

                if function_update.response == FunctionUpdateResponse.FUNCTION_UPDATE_SUCCESS:
                    logging.info(f"Function [{function_update.function_registration.details.id}] registration was successful")
                elif function_update.response == FunctionUpdateResponse.FUNCTION_UPDATE_UNAUTHORIZED:
                    logging.warning(f"Function [{function_update.function_registration.details.id}] signature check failed!")
                elif function_update.response == FunctionUpdateResponse.FUNCTION_UPDATE_UNAVAILABLE:
                    logging.warning(f"Function [{function_update.function_registration.details.id}] is no longer registered")
                else:
                    logging.warning(f"Function [{function_update.function_registration.details.id}] registration failed!")

                if function_update.HasField("response_description"):
                    logging.warning(function_update.response_description)

    def _get_function_instance(self, call: FunctionCall) -> typing.Optional[OpenOBDFunction]:
        """
        Returns an instance of the OpenOBDFunction class, implementing the function specified by the given function call object.
        :param call: Object identifying the function to call (id) and the session to use (session_info)
        :return: OpenOBDFunction instance
        """
        return self.function_executor.instantiate_function_from_uuid(id=call.id,
                                                                     openobd_session=OpenOBDSession(call.session_info, grpc_factory=self.grpc_factory),
                                                                     function_broker=self.function_broker)

    def _call(self, function: OpenOBDFunction):
        """
        Runs the given OpenOBDFunction on the current executor
        :param function:
        :return:
        """
        self.function_executor.run_function(function)

    def _available(self, call: FunctionCall):
        """
        Checks whether a function is available for the given function call. Raises an error if the function is not available
        :param call:
        :return: void
        """
        self.function_executor.instantiate_function_from_uuid(id=call.id,
                                                              openobd_session=OpenOBDSession(call.session_info, grpc_factory=self.grpc_factory),
                                                              function_broker=self.function_broker,
                                                              dry_run=True)







from collections.abc import Iterator

from openobd_protocol.Function.Messages import Function_pb2 as grpcFunction
from openobd_protocol.FunctionBroker.Messages import FunctionBroker_pb2 as grpcFunctionBroker
from openobd_protocol.SessionController.Messages import SessionController_pb2 as grpcSessionController

from openobd.core.arguments import ConnectionArguments
from openobd.core.exceptions import raises_openobd_exceptions
from openobd.core.session_controller import OpenOBDSessionController
from openobd.core.session import OpenOBDSession
from openobd.core.function_broker import OpenOBDFunctionBroker


class SingletonMeta(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class OpenOBD(metaclass=SingletonMeta):

    def __init__(self, arguments: ConnectionArguments = None):
        """
        A singleton that allows for the starting and managing of openOBD sessions using the provided Partner API
        credentials. The credentials are retrieved from the arguments.
        """

        # For backwards compatibility, we allow instantiating this class without arguments. In that case,
        # the environment variables are used to configure the connection
        arguments = arguments if arguments else ConnectionArguments.from_environment_variables()
        self.session_controller = OpenOBDSessionController(arguments)
        self.function_broker = OpenOBDFunctionBroker(arguments)

    @raises_openobd_exceptions
    def start_session_on_ticket(self, ticket_id: str) -> OpenOBDSession:
        """
        Starts an openOBD session on the given ticket.

        :param ticket_id: the ticket number (or identifier) on which a session should be started.
        :return: an OpenOBDSession object representing the started session, which can be used to make gRPC calls.
        """
        response = self.session_controller.start_session_on_ticket(grpcSessionController.TicketId(value=ticket_id))
        return OpenOBDSession(response)

    @raises_openobd_exceptions
    def start_session_on_connector(self, connector_id: str) -> OpenOBDSession:
        """
        Starts an openOBD session on the given connector.

        :param connector_id: the UUID of the connector on which a session should be started.
        :return: an OpenOBDSession object representing the started session, which can be used to make gRPC calls.
        """
        response = self.session_controller.start_session_on_connector(grpcSessionController.ConnectorId(value=connector_id))
        return OpenOBDSession(response)

    @raises_openobd_exceptions
    def get_session(self, session_id: grpcSessionController.SessionId) -> grpcSessionController.SessionInfo:
        """
        Retrieves the requested openOBD session. Raises an OpenOBDException if the session does not exist.

        :param session_id: the identifier of the session to be retrieved.
        :return: a SessionInfo object representing the requested session.
        """
        return self.session_controller.get_session(session_id)

    @raises_openobd_exceptions
    def interrupt_session(self, session_id: grpcSessionController.SessionId) -> grpcSessionController.SessionInfo:
        """
        Forcefully closes the given openOBD session. This changes the session's state to "interrupted" and prevents
        further communication with the session.

        :param session_id: the identifier of the session to be interrupted.
        :return: a SessionInfo object representing the interrupted session.
        """
        return self.session_controller.interrupt_session(session_id)

    @raises_openobd_exceptions
    def get_session_list(self) -> grpcSessionController.SessionInfoList:
        """
        Retrieves all (recently) active openOBD sessions for this partner.

        :return: a SessionInfoList object containing an iterable of SessionInfo objects under its "sessions" attribute.
        """
        return self.session_controller.get_session_list()

    @raises_openobd_exceptions
    def open_function_stream(self, function_update_messages: Iterator[grpcFunctionBroker.FunctionUpdate]) -> Iterator[grpcFunctionBroker.FunctionUpdate]:
        """
        Opens a stream in which functions can be made available for function callers, and through which function calls
        will be forwarded.

        :param function_update_messages: a FunctionUpdate message for each function that needs to be registered.
        :return: FunctionUpdate messages containing acknowledgements of registrations, pings, and function calls.
        """
        return self.function_broker.open_function_stream(function_update_messages)

    @raises_openobd_exceptions
    def run_function(self, function_id: str, session_info: grpcSessionController.SessionInfo) -> grpcFunctionBroker.FunctionUpdate:
        """
        Execute a function that has been registered by a function launcher.

        :param function_id: the UUID of the function to call.
        :param session_info: the session to run the function in.
        :return: a FunctionUpdate object defining whether the function has been successfully launched.
        """
        return self.function_broker.run_function(grpcFunctionBroker.FunctionCall(id=function_id, session_info=session_info))

    @raises_openobd_exceptions
    def get_function_registration(self, function_id: str) -> grpcFunctionBroker.FunctionRegistration:
        """
        Retrieves information about the requested function. For instance, whether the function is online or not.

        :param function_id: the UUID of the function to request info on.
        :return: a FunctionRegistration object containing details on the requested function.
        """
        return self.function_broker.get_function_registration(grpcFunction.FunctionId(value=function_id))

    @raises_openobd_exceptions
    def generate_function_signature(self) -> grpcFunctionBroker.FunctionSignature:
        """
        Generates a new function ID and signature, which are used when registering an openOBD function.

        :return: a FunctionSignature object containing a new function ID and its corresponding signature.
        """
        return self.function_broker.generate_function_signature()





import logging
from openobd import FunctionExecutor, OpenOBDFunctionBroker
from openobd.cli import CliParser
from openobd.core.session_builder import SessionBuilder
from openobd.functions.function_launcher import FunctionLauncher
from openobd.log import initialize_logging

if __name__ == "__main__":

    cli = CliParser().parse()
    initialize_logging(cli.log_level)

    logging.debug(f"Using cluster {cli.connection_arguments.cluster_id} and grpc server {cli.connection_arguments.grpc_host}:{cli.connection_arguments.grpc_port}")
    executor = FunctionExecutor(cli.executor_arguments)

    if cli.is_command_run():
        SessionBuilder(cli.connection_arguments, cli.session_arguments, executor).run(file_name = cli.file)

    elif cli.is_command_serve():
        function_broker = OpenOBDFunctionBroker(cli.connection_arguments)
        FunctionLauncher(function_broker, executor).serve()



import logging
from logging import Logger
from typing import Optional
from abc import ABC, abstractmethod

from openobd_protocol.FunctionBroker.Messages.FunctionBroker_pb2 import FunctionRegistration, FunctionRegistrationState, \
    FunctionVisibility
from openobd_protocol.Function.Messages.Function_pb2 import FunctionDetails, VariableList, ContextType, Product, Vehicle
from openobd_protocol.Messages.Empty_pb2 import EmptyMessage
from openobd_protocol.Session.Messages.ServiceResult_pb2 import ServiceResult, Result

from openobd.core.exceptions import OpenOBDException
from openobd.core.session_token_handler import SessionTokenHandler
from openobd.core.session import OpenOBDSession
from openobd.core.function_broker import OpenOBDFunctionBroker
from openobd.functions.function_context_handler import FunctionContextHandler
from openobd.functions.function_context_variables import ContextVariables
from openobd.functions.message import pack, get_message_classes_from_function_signature, unpack_variable_list
from openobd.log import ScriptLogHandler

'''When this key is found in the function results, it means we have a single return value (no dict)'''
FUNCTION_SINGLE_RETURN_VALUE = "<FUNCTION_SINGLE_RETURN_VALUE>"

class OpenOBDFunction(ABC):
    id = ""
    version = "<undef>"
    name = ""
    description = ""
    signature = ""

    # List of product uuids that this function supports. Leave empty for generic functions
    products = []                    # type: list[str]

    # List of vehicle ids that this function supports. Leave empty for generic functions
    vehicles = []                    # type: list[str]

    dashboard = True

    openobd_session = None          # type: OpenOBDSession
    session_variables = None        # type: ContextVariables
    function_variables = None       # type: ContextVariables
    connection_variables = None     # type: ContextVariables
    result = None                   # type: ServiceResult
    message_classes = []
    _signature_classes = []
    logger = None                   # type: Logger
    _log_handler = None             # type: ScriptLogHandler

    def __init__(self):
        self.function_broker = None
        self.result = ServiceResult(result=[Result.RESULT_SUCCESS])
        self._context_finished_ = None
        self._function_registration = None
        self.logger = logging.getLogger(type(self).__name__)

    '''
    Optionally you can pass a OpenOBDFunctionBroker (when non-standard configuration is needed)
    '''
    def initialize(self, openobd_session: OpenOBDSession, function_broker: Optional[OpenOBDFunctionBroker] = None):
        try:
            self.openobd_session = openobd_session
            self.function_broker = function_broker
            self._initialize_function_details()
            self._initialize_function_registration()
            self._context_finished_ = False
            self._signature_classes = get_message_classes_from_function_signature(self.run)

            '''
            When no message classes are defined (protobuf messages), we at least handle the message classes
            that appear in the function signature of the run() function
            '''
            if len(self.message_classes) == 0:
                self.message_classes = self._signature_classes

            '''Initialize variable contexts'''
            self._initialize_variable_contexts(self.message_classes)

        except OpenOBDException as e:
            logging.error(f"Failed to initialize openOBD function: {e}")
            self._function_registration = None
            self._context_finished_ = True

    def _initialize_variable_contexts(self, message_classes = None):
        self.session_variables = ContextVariables(self.openobd_session, ContextType.GLOBAL_CONTEXT, message_classes)
        self.connection_variables = ContextVariables(self.openobd_session, ContextType.CONNECTION_CONTEXT, message_classes)
        self.function_variables = ContextVariables(self.openobd_session, ContextType.FUNCTION_CONTEXT, message_classes)

    def _initialize_function_registration(self):
        if self.dashboard:
            visibility = FunctionVisibility.FUNCTION_VISIBILITY_DASHBOARD
        else:
            visibility = FunctionVisibility.FUNCTION_VISIBILITY_INTERNAL

        self._function_registration = FunctionRegistration(
            details=self._function_details,
            state=FunctionRegistrationState.FUNCTION_REGISTRATION_STATE_ONLINE,
            signature=self.signature,
            visibility=visibility
        )

    @abstractmethod
    def run(self, **kwargs) -> Optional[dict]:
        pass

    def get_function_registration(self):
        return self._function_registration

    def get_function_id(self):
        return self.id

    def _initialize_function_details(self, id=None, version=None, name=None, description=None, signature=None):
        self._function_details = None

        '''Check id presence, if not present we generate a fresh one'''
        if not id is None:
            self.id = id
        elif len(self.id) == 0:
            if self.function_broker is None:
                raise OpenOBDException("Function broker has not been initialized")

            function = self.function_broker.generate_function_signature()
            self.id = function.id
            self.signature = function.signature

        '''Check signature presence'''
        if not signature is None:
            self.signature = signature
        elif len(self.signature) == 0:
            raise OpenOBDException("A function signature is required in order to serve a function on the network!")

        '''Initialize (version, name, description) correctly'''
        if not version is None:
            self.version = version

        '''Only overwrite name with class name when no custom name is provided'''
        if not name is None:
            self.name = name
        elif len(self.name) == 0:
            self.name = self.__class__.__name__

        if not description is None:
            self.description = description

        self._function_details = FunctionDetails(id=self.id,
                                                 version=self.version,
                                                 name=self.name,
                                                 description=self.description,
                                                 products=[Product(id=productId) for productId in self.products],
                                                 vehicles=[Vehicle(id=vehicleId) for vehicleId in self.vehicles])

    def __enter__(self):
        if self._context_finished_:
            ''' We cannot continue when the constructor failed '''
            raise OpenOBDException("Failed to construct openOBD session")

        self._authenticate()
        return self

    # Authenticate against the openOBD session for this function
    def _authenticate(self):
        try:
            # Start the SessionTokenHandler to ensure the openOBD session remains authenticated
            SessionTokenHandler(self.openobd_session)
        except OpenOBDException as e:
            logging.error(f'Activating openOBD session failed: {e}')
            self._context_finished_ = True
            raise

    def __is_active__(self):
        return not self._context_finished_

    def __del__(self):
        self.__exit__(None, None, None)

    def interrupt(self):
        self.__exit__(OpenOBDException(f"Interrupt"), f"Function [{self.id}] has been interrupted!", None)

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type:
            if exc_type == OpenOBDException:
                description = f'Request failed: {exc_value}'
                logging.error(description)
            else:
                description = f"Something unexpected occurred [{exc_value}]"
                # if traceback is not None:
                #     logging.warning(traceback.format_exc())
            logging.error(description)
            self.result = ServiceResult(result=[Result.RESULT_FAILURE_RUNTIME_ERROR], description=description)

        # Remove the log handler that might have been initialized
        self.disable_logging_to_scriptlog()

        # Make sure we finish the context once
        if not hasattr(self, '_context_finished_'):
            # Something wrong with initialization, also no cleanup possible
            return

        if self._context_finished_:
            return

        self._context_finished_ = True
        if self.openobd_session is not None:
            # Check if result is set
            if self.result is None:
                self.result = ServiceResult(result=[Result.RESULT_FAILURE])

            # Check if result is of correct type
            if not isinstance(self.result, ServiceResult):
                self.result = ServiceResult(result=[Result.RESULT_FAILURE_RUNTIME_ERROR],
                                            description="Incorrect result type used")

            self.openobd_session.finish(self.result)
            self.openobd_session = None

    def function_call(self, function: type | str, **kwargs):
        """
        Calls a function with the given arguments and waits until it is finished before returning its results.

        :param function: either a reference to the function's class or the ID of the function.
        :param kwargs: arguments to pass to the function.
        :return: any results set by the called function.
        """

        '''Determine the function ID depending on the given argument type'''
        if isinstance(function, type) and issubclass(function, OpenOBDFunction):
            function_id = function.id
            function_class = function
        elif isinstance(function, str):
            function_id = function
            function_class = None
        else:
            raise OpenOBDException(f"Function not found: {function}")

        '''Create fresh context'''
        function_context = FunctionContextHandler(self.openobd_session)

        '''Add eventual function arguments to the context'''
        self._set_function_arguments(**kwargs)

        '''Hand over the function handle to an executor (via the function broker)'''
        function_context.run_function(function_id, self.function_broker)

        '''Return eventual results'''
        return self.get_results(function_class)

    def _set_function_arguments(self, **kwargs):
        for key, value in kwargs.items():
            var = pack(value)
            logging.debug(f"Setting function argument [{key}] to [{var}]")
            self.openobd_session.set_function_argument(key, var)

    '''Return the arguments as a dictionary'''
    def get_function_arguments(self) -> dict:
        function = self.openobd_session.function
        argument_list = function.getFunctionArgumentList(request=EmptyMessage(),
                                                         metadata=self.openobd_session.metadata())  # type: VariableList

        return unpack_variable_list(argument_list, self._signature_classes)

    def set_result(self, key, value):
        self.openobd_session.set_function_result(key=key, value=pack(value))

    '''Return the results as a dictionary'''
    def get_results(self, function_class=None):
        function = self.openobd_session.function
        result_list = function.getFunctionResultList(request=EmptyMessage(),
                                                     metadata=self.openobd_session.metadata())  # type: VariableList

        '''Convert to dictionary and unpack eventual signature protobuf Message classes'''
        message_classes = get_message_classes_from_function_signature(function_class.run) if function_class is not None else None
        variables = unpack_variable_list(result_list, message_classes)

        '''Check if there has been a return value at all, if not we return None'''
        if len(variables) == 0:
            return None

        '''Check if there was a single return value (the other remaining option is a dict)'''
        for key in variables.keys():
            if key == FUNCTION_SINGLE_RETURN_VALUE:
                '''We have a single result value, return immediately'''
                return variables[key]

        '''A dict was used as return value of the function, return this dict'''
        return variables

    def enable_logging_to_scriptlog(self, level=None):
        '''Check if we have a requested log level'''
        if isinstance(level, str):
            self.logger.setLevel(level)

        if self._log_handler:
            '''Initialize only once'''
            return

        self._log_handler = ScriptLogHandler(self.openobd_session)
        self.logger.addHandler(self._log_handler)

    def disable_logging_to_scriptlog(self):
        if not self._log_handler:
            '''Only remove when initialized'''
            return

        self.logger.removeHandler(self._log_handler)
        self._log_handler.close()
        self._log_handler = None



import importlib
import os
import inspect
import logging
import pathlib
import typing

from openobd_protocol.FunctionBroker.Messages.FunctionBroker_pb2 import FunctionRegistration

from openobd.functions.message import is_protobuf, pack
from openobd.core.exceptions import OpenOBDException
from openobd.core.function_broker import OpenOBDFunctionBroker
from openobd.core.arguments import ExecutorArguments
from openobd.functions.composition import OpenOBDComposition
from openobd.functions.function import OpenOBDFunction, FUNCTION_SINGLE_RETURN_VALUE


class FunctionExecutor:

    def __init__(self, arguments: ExecutorArguments):
        self.arguments = arguments
        self.functions = {}

    def load_modules(self, function_broker: OpenOBDFunctionBroker):
        if self.arguments.unique_function_ids:
            logging.info("[UNIQUE] Generating fresh function id's for all loaded functions")

        modules = self.arguments.modules
        if modules is not None:
            for module in modules:
                if not os.path.isdir(module):
                    raise Exception(f"Given module directory does not exist in working path {os.getcwd()}: {module}")

                for root, dirs, files in os.walk(module):
                    for file in files:
                        if file.endswith('.py') and file != '__init__.py':
                            module_name = ".".join(pathlib.Path(root, file).with_suffix("").parts)
                            try:
                                importlib.import_module(module_name)
                                self.load_function(f"{module_name}", function_broker)
                            except ModuleNotFoundError as e:
                                logging.warning(f"Failed to import {module_name}: {e}")

    def load_function(self, module: str, function_broker: OpenOBDFunctionBroker) -> typing.Type[OpenOBDFunction]:
            logging.info(f"Loading: [{module}]")
            function_reference = None
            unique_function_ids = self.arguments.unique_function_ids
            function_name_prefix = self.arguments.prefix

            try:
                mod = importlib.import_module(module)
                for name, obj in inspect.getmembers(mod, inspect.isclass):
                    # Only include classes defined in this module (not imported ones)
                    if obj.__module__ == module:
                        cls = getattr(mod, name)

                        '''Check for fingerprint of openOBD function or composition'''
                        if hasattr(cls, 'id') and hasattr(cls, 'signature') and hasattr(cls, 'name'):
                            if unique_function_ids:
                                '''Ensure fresh registration'''
                                setattr(cls, 'id', '')
                                setattr(cls, 'signature', '')

                            '''Prefix adjustment is only possible when we know that it is really a function or composition'''
                            if unique_function_ids and function_name_prefix:
                                setattr(cls, 'name', f"[{function_name_prefix}] {getattr(cls, 'name')}")

                            # Initialize the function class (Either a function or a composition)
                            try:
                                function = cls()
                                function.initialize(openobd_session=None, function_broker=function_broker)
                            except TypeError as e:
                                logging.debug(f"Not a function or composition class: {name}", exc_info=e)
                                continue

                            if isinstance(function, OpenOBDComposition) or isinstance(function, OpenOBDFunction):
                                logging.info(f"Initializing function: [{function.id}] {function.name}")

                                '''Initialize the class in memory with the registered values'''
                                setattr(cls, 'id', function.id)
                                setattr(cls, 'signature', function.signature)
                            else:
                                '''Other classes might be helper classes, we do not need to initialize them'''
                                continue

                            function_registration = function.get_function_registration()  # type: FunctionRegistration
                            function_id = function_registration.details.id
                            if function_id in self.functions and self.functions[function_id][1] != module:
                                logging.warning(f"Function ID [{function_id}] is already used by another function")
                            self.functions[function_id] = (function_registration, module, name)
                        else:
                            '''When the expected fingerprint is not present, we do not initialize'''
                            continue

                        '''
                        First function found in a file is considered the 'main' function
                        General advice is to just define one function per file (to prevent any misunderstanding)
                        '''
                        if function_reference is None:
                            function_reference = cls
            except (ModuleNotFoundError, ValueError) as e:
                logging.critical(f"The module could not be found {module}: {e}")
            except Exception as e:
                logging.error(f"Problem loading functions from module {module}: {e}")
                raise e

            return function_reference

    def get_function_registrations(self):
        function_registrations = []
        for function_id in self.functions.keys():
            function_registrations.append(self.functions[function_id][0])
        return function_registrations

    def instantiate_function_from_uuid(self, id, openobd_session, function_broker=None, dry_run=False) -> typing.Optional[OpenOBDFunction]:
        if id not in self.functions:
            raise OpenOBDException(f"Function {id} unknown!")

        (function_registration, full_module_name, name) = self.functions[id]
        module = importlib.import_module(full_module_name)

        if dry_run:
            return None

        # Instantiate function class
        cls = getattr(module, name)

        logging.debug(f"Instantiate function [{id}]: {name}")
        function_instance = cls()
        function_instance.initialize(openobd_session=openobd_session, function_broker=function_broker)
        return function_instance

    @staticmethod
    def run_function(function: OpenOBDFunction, **kwargs):
        results = None
        with function as f:
            '''When kwargs are passed explicitly, use the kwargs from function call'''
            if len(kwargs) == 0:
                '''Load eventual arguments to pass to the run() function from session'''
                kwargs = f.get_function_arguments()

            '''Make the call to the run() function with eventual arguments'''
            results = f.run(**kwargs)
            '''Check if eventual results need to be set (only dict can be result)'''
            if isinstance(results, dict):
                for key in results.keys():
                    f.set_result(key, pack(results[key]))
            elif isinstance(results, str):
                f.set_result(FUNCTION_SINGLE_RETURN_VALUE, results)
            elif is_protobuf(results):
                f.set_result(FUNCTION_SINGLE_RETURN_VALUE, pack(results))
            elif results is not None:
                f.set_result(FUNCTION_SINGLE_RETURN_VALUE, str(results))

        return results

    def __str__(self):
        function_list = " Functions ".center(80, "-") + "\n"
        if len(self.functions) == 0:
            function_list += " " * 10 + "<none>\n"
        else:
            for function_id in self.functions.keys():
                function_registration = self.functions[function_id][0]
                function_list += "\n"
                function_list += f"             Id: {function_registration.details.id}\n"
                function_list += f"      Signature: {function_registration.signature}\n"
                function_list += f"           Name: {function_registration.details.name}\n"
                function_list += f"        Version: {function_registration.details.version}\n"
                function_list += f"    Description: {function_registration.details.description}\n"
            function_list += "\n"
        function_list += "-" * 80
        return function_list

